/* SPDX-License-Identifier: MIT */
#ifndef AMDGPU_BACKPORT_H
#define AMDGPU_BACKPORT_H

#include <linux/version.h>
#include <linux/kernel.h>
#include <kcl/kcl_moduleparam.h>
#include <kcl/kcl_types.h>
#include <kcl/kcl_idr.h>
#include <kcl/kcl_kernel.h>
#include <kcl/kcl_compiler_attributes.h>
#include <kcl/kcl_dma_mapping.h>
#include <kcl/kcl_dma-buf-map.h>
#include <kcl/kcl_iosys-map.h>
#include <kcl/kcl_eventpoll.h>
#include <kcl/kcl_i2c.h>
#include <kcl/kcl_video.h>
#include <kcl/kcl_acpi.h>
#include <kcl/kcl_workqueue.h>
#include <kcl/kcl_backlight.h>
#include <kcl/kcl_device.h>
#include <kcl/backport/kcl_fs.h>
#include <kcl/backport/kcl_kthread_backport.h>
#include <kcl/kcl_list.h>
#include <kcl/backport/kcl_io_backport.h>
#include <kcl/backport/kcl_uaccess_backport.h>
#include <kcl/kcl_seq_file.h>
#include <kcl/kcl_seqlock.h>
#include <kcl/kcl_mmu_notifier.h>
#include <kcl/kcl_preempt.h>
#include <kcl/kcl_suspend.h>
#include <kcl/backport/kcl_pci_backport.h>
#include <kcl/kcl_timekeeping.h>
#include <kcl/backport/kcl_mm_backport.h>
#include <kcl/kcl_pagemap.h>
#include <kcl/kcl_pm.h>
#include <kcl/backport/kcl_fence_backport.h>
#include <kcl/backport/kcl_device_cgroup_backport.h>
#include <kcl/backport/kcl_hmm.h>
#include <kcl/backport/kcl_migrate.h>
#include <kcl/kcl_mn.h>
#ifdef HAVE_DRM_DRIVER_GEM_PRIME_RES_OBJ
#include <linux/reservation.h>
#endif
#include <kcl/backport/kcl_drm_cache_backport.h>
#include <kcl/kcl_task_barrier.h>
#include <kcl/backport/kcl_drm_print.h>
#include <kcl/backport/kcl_drm_backport.h>
#include <kcl/backport/kcl_drm_gem.h>
#include <kcl/backport/kcl_drm_prime.h>
#include <kcl/backport/kcl_drm_vma_manager_backport.h>
#include <kcl/backport/kcl_drm_fb.h>
#include <kcl/backport/kcl_drm_drv.h>
#include <kcl/kcl_drm_modes.h>
#include <kcl/kcl_drm_connector.h>
#include <kcl/backport/kcl_drm_dp_helper_backport.h>
#include <kcl/backport/kcl_drm_dp_mst_helper_backport.h>
#include <kcl/backport/kcl_drm_atomic_helper_backport.h>
#include <kcl/kcl_drm_hdcp.h>
#include <kcl/kcl_moduleparam.h>
#include <kcl/kcl_drm_prime.h>
#include <kcl/kcl_acpi_table.h>
#include <kcl/kcl_drm_fourcc.h>
#include <kcl/kcl_debugfs_inode.h>
#include <kcl/kcl_sysfs_emit.h>
#include <kcl/kcl_bitops.h>
#include <kcl/kcl_sched_mm.h>
#include <kcl/backport/kcl_numa_backport.h>
#include <kcl/kcl_mmzone.h>
#include "kcl/kcl_amdgpu_drm_fb_helper.h"
#include "kcl/kcl_amdgpu.h"
#include "kcl/kcl_amdgpu_drm_drv.h"
#include "kcl/kcl_amdgpu_drm_gem.h"
#include "kcl/kcl_drm_gem_ttm_helper.h"
#include "kcl/kcl_mce.h"
#include "kcl/kcl_drm_aperture.h"
#include <kcl/kcl_intel_family.h>
#include <kcl/kcl_drm_simple_kms_helper.h>
#include <kcl/backport/kcl_bitmap.h>
#include <kcl/kcl_capability.h>
#include <kcl/kcl_fdtable.h>
#include <kcl/kcl_bitmap.h>
#include <kcl/kcl_dma_fence_chain.h>
#include <kcl/kcl_pci.h>
#include <kcl/kcl_cc_platform.h>
#include <kcl/kcl_module.h>
#include <kcl/kcl_stddef.h>
#include <kcl/kcl_mce.h>
#include <kcl/kcl_hypervisor.h>
#include <kcl/backport/kcl_workqueue_backport.h>
#include <kcl/kcl_string_helpers.h>
#include <kcl/kcl_cpumask.h>
#include <kcl/kcl_drm_modeset_lock.h>
#include <kcl/kcl_drm_dsc_helper.h>
#include <kcl/kcl_highmem-internal.h>
#include <kcl/kcl_dynamic_debug.h>
#include <kcl/kcl_drm_plane_helper.h>
/*
 * v5.13-rc3-1669-gba6cd766e0bf
 * ("drm/plane: Move drm_plane_enable_fb_damage_clips into core")
 * move drm_plane_enable_fb_damage_clips() to drm_planer.h.
 * include drm_damage_helper.h to fix the missing function declaration for legacy kernel.
 */
#include <drm/drm_damage_helper.h>
#include <kcl/backport/kcl_drm_probe_helper.h>
#include <kcl/kcl_rbtree.h>
#include <kcl/kcl_delay.h>
#include <kcl/kcl_apple-gmux.h>
#include <kcl/backport/kcl_drm_edid.h>
#include <kcl/kcl_debugfs.h>
#include <kcl/kcl_irqdesc.h>
#include <kcl/kcl_pci_ids.h>
#include <kcl/kcl_drm_suballoc.h>
#include <kcl/kcl_drm_edid.h>
#include <kcl/kcl_drm_file.h>
#include <kcl/kcl_drm_exec.h>
#include <kcl/kcl_local64.h>
#include <kcl/kcl_class.h>
#include <kcl/kcl_build_bug.h>
#include <kcl/kcl_drm_dp.h>
#include <kcl/kcl_drm_writeback.h>
#include <kcl/kcl_drm_client.h>
#include <kcl/kcl_drm_mode.h>
#include <kcl/kcl_acpi_amd_wbrf.h>
#include <kcl/kcl_units.h>
#include <kcl/kcl_dma-buf.h>
#include <kcl/backport/kcl_drm_exec.h>
#endif /* AMDGPU_BACKPORT_H */
